(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/AbsynPrint.sml,v
   Comment: PrettyPrinting Absyn elements for testing and debugging
   Release: scope-0-16_final

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   TODO: the module had been created ad hoc for testing and debugging in 
         interactive environment (using pretty-printers hooked into the
	 compiler). Interestingly it started to be used also in visualization
	 of sttecharts (also a debugging-oriented module initially) and now
	 it happens that it processes quite larger chunks of data. If the
	 efficiency is a bottleneck then this needs to be
	 rewritten/redesigned in CPS style at least + minimize number of
	 string concatanations (2003/05/06).

   AbsynPrint.sml,v 1.9 2003/07/07 23:53:15 wasowski Exp
*)
structure AbsynPrint = struct

open Absyn 
infixr 0 $
infix 9 sub

structure S = Symbol
val all = List.all

(* some combinators *)

fun lift_pr  (f :'a -> string) :'a -> unit = fn a => print (f a)
fun lift_ln  (f :'a -> 'b    ) :'a -> 'b = fn a => f a before print "\n"

(* and The real thing *)

fun sprEBind (sy :symbol, [] :symbol list) : string = S.name sy
  | sprEBind (sy :symbol, pa :symbol list) : string = 
      String.concat (Symbol.name sy ::  
	Typeset.c_params (map Symbol.name pa) [])

fun sprTrigger (Event eb) : string = sprEBind eb
  | sprTrigger (EventGroup g) = Symbol.name g

fun sprStateor  (ORState  s) = S.name (#name s)
fun sprStateand (ANDState s) = S.name (#name s)
fun sprState (s :'aor state) = lift_orand sprStateor sprStateand s

fun sprId (s :stateid) :string = Symbol.name s
fun sprIds (ids :stateid list) = 
      String.concat $ Typeset.join_blanks(map sprId ids) []
val prIds = lift_pr sprIds

val sprInt = Int.toString
fun sprInts (is :int list) = 
      String.concat $ Typeset.join_blanks(map sprInt is) []
val prInts = lift_pr sprInts

(* guard and action supressed *)

fun sprPos (ss :stateid list) : string =
      String.concat("POS(" ::Typeset.join_blanks(map S.name ss) [")"])

fun sprNeg (ss :stateid list) : string =
      String.concat( "NEG(" ::Typeset.join_blanks(map S.name ss) [")"])

val prPos = lift_pr sprPos
val prNeg = lift_pr sprNeg
val prPosln = lift_ln prPos
val prNegln = lift_ln prNeg

fun sprCnst (CstI i) = Int.toString i
  | sprCnst (CstR r) = Real.toString r
  | sprCnst (CstC c) = implode [#"'",c,#"'"]

fun sprExp (Cst c) = sprCnst c
  | sprExp (Access a) = sprAccess a
  | sprExp (Prim1 (ope,e)) = String.concat 
	[CodGenC.dump_Prim1 ope, "(",sprExp e,")"]
  | sprExp (Prim2 (ope, e1,e2)) = String.concat ["(", sprExp e1, ")", 
	CodGenC.dump_Prim2 ope, "(", sprExp e2, ")"]
  | sprExp (Call  (ou,es)) = String.concat 
	(S.name ou::"("::Typeset.join_commas (map sprExp es) [")"])
  | sprExp (After (e1,e2)) = String.concat
	["(", sprExp e1, "),(", sprExp e2, ")"]

and sprAccess (AccessVar  v   ) = S.name v
  | sprAccess (AccessArr (v,e)) = 
	String.concat [S.name v, "[", sprExp e, "]"]

fun sprCmd (Expr (Call (ou,[]))) = S.name ou
  | sprCmd (Expr e)      = sprExp e
  | sprCmd (Assgn (a,e)) = String.concat [sprAccess a, "=", sprExp e]

fun sprGuard lst = 
      String.concat (Typeset.join_blanks (map sprExp lst) [])

fun sprSig (e,[])= "^" ^ S.name e
  | sprSig (e, _)= raise Fail "#200305060:Not implemented"

fun sprAction (c, sigs) = String.concat 
			(Typeset.join_blanks (map sprCmd c)  
			(" "::Typeset.join_blanks (map sprSig sigs) []))
      
fun sprTran ( {trigger,scope,pos,neg,guard,action,targets,annotation}:'atr tran) =
      String.concat (
      "[" :: S.name scope :: "] " :: sprTrigger trigger :: " : " 
	:: sprPos pos :: " " :: sprNeg neg :: " -> " 
		:: Typeset.join_blanks(map S.name targets) [])
		
fun prTran  t  = lift_pr sprTran t
fun prTranln t = lift_ln prTran  t

end
