(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenC1_LR.sml,v
   Comment: intermediate representation (mostly to be used by CodGenC1.
   Release: scope-0-16_final 
  
     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   All types used in intermediate representation are defined here.
   Data types only. No logics defined here (except from trivialities).

   CodGenC1_LR.sml,v 1.28 2003/07/21 23:01:22 wasowski Exp
*)

structure CodGenC1_LR = struct

open LR


(* number of markers (model dependent). We keep it constant but easily
   updateable by hand. To make it really dynamic and model dependent it
   should be moved to one of model representation/information records.
   (FIXME: this would be a nice thing to do in production version) *)

val markno :int = 5

val always_ref :guardref = markno (* must live here bcs depndent on markno *)
val null_actref :int    = markno


(* information about exit types present *)

type exitmode = { pure : bool, impure : bool }

fun use_pure   mode = if #pure mode then mode
		      else { pure=true, impure= #impure mode }

fun use_impure mode = if #impure mode then mode
		      else { pure= #pure mode, impure = true }


(* system in intermediate representation *)

type lrsystem = { 

	name	 :string,		(* system name used for generating file name *)
	fname    :string,		(* name of the source file           *)
	
	hor	 :lratom list,		(* or part of hierarchy		     *)
	hand	 :lratom list,		(* and part of hierarchy	     *)
	root     :symbol,		(* name of topstate                  *)
	tenv     :tenv,			(* transitions array		     *)
	
	history  :lratom list,		(* list of local history vectors     *)
	signals  :signals,		(* list of triggering parts of acts  *)
	
	actions  :actions,		(* unique action bodies (ordered)    *)
	guards	 :guards,		(* unique guard bodies (ordered)     *)

	evtypes  :evtypes,		(* Unmodified from Absyn.sml         *)
	vartypes :vartypes,		(* Unmodified from Absyn.sml	     *)
	venv     :venv,			(* Unmodified from Absyn.sml	     *)
	cenv     :const env,		(* Unmodified from Absyn.sml         *)

	queuelen :int option,		(* only SOME if queue is used        *)
	statew	 :int,			(* upper bound of state configuration*) 
					(* size				     *)
	targetw  :int,			(* upper bound for number of targets *)
					(* in any single entering	     *)
	orsts    :int,			(* number of or-states		     *)
	andsts   :int,			(* number of and-states              *)
	exitmode :exitmode		(* exit mode used		     *)
}	 
 
(* scheduling related types *)

(* state addresses with automatic compression *)

datatype stdict = Direct   of int env
		| Indirect of int env * int list * int
		(* names to entries in dict, dict, no cells in dict *)

fun lookup_stdict (Direct env) = SE.lookup env
  | lookup_stdict (Indirect (env,arr,size)) = SE.lookup env

(* for debugging purposed *)

fun toList_stdict (Direct env) = SE.toList env
  | toList_stdict (Indirect (env,arr,size)) = SE.toList env

(* type of data produced by scheduler *)

type scheduling = {

	eenv	  :eenv,	(* map event names to indices		     *)
	hor_dict  :stdict,	(* map or-states to addresses in hor (mconf) *)
	hand_dict :stdict,	(* map andstates to addresses in hand        *)
	tranidx   :tranidx      (* map events to position in transarr        *)
	
}

type lrmodel = (lrsystem * lrsizes) list

end

