(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenC1_schedule.sml,v
   Comment: Scheduler for CodGenC1
   Release: scope-0-16_final 

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   The task of the scheduler is to resolve all string references in
   intermediate representation code generated by CodGenC1. Scheduler takes
   lrsystem and returns scheduling data for it.
   
   CodGenC1_schedule.sml,v 1.18 2003/07/21 15:21:17 wasowski Exp *)

structure CodGenC1_schedule = struct

(* some interface *)

open Haskellish Err Say CodGenC1_LR CodGenC_schedule

infixr 1 $ 

structure SE = SymbolEnv


	    

(* get number of cells used for entry/exit actions. NOTE: this must be
   in synchro with corresponding pack_enex function in back end *)

fun measure_enex (ac_size :int) (hand_size :int) (en :int, ex :int) :int =
	if en = null_actref andalso ex = null_actref then 0
	else if 2 * ac_size <= hand_size then 1
	else if ex = null_actref then measure1 ac_size hand_size
	else measure2 (ac_size,ac_size) hand_size

	

(* build a dictionary for accessing states       *)
(* the same function for or-state and and-states *)
(* offset is the number of cells to reserve in   *)
(* the begining of dictionary to avoid clashes   *)
(* with markers of LR                            *)

(* TODO: once indirect index is needed we could  *)
(* avoid using prefix offset in hor and hand -   *)
(* these are not used as state indices anymore.  *)
(* brings save of something like 4 bytes         *)

fun build_dict confset (arr :lratom list) (offset :int) 
	(sizes :lrsizes) (tgt_size :int) :stdict =
	let val g = measure_enex (#actionref sizes) (#orstref sizes)
	    fun f lst n max dict = case lst of 
			  [] => (max, dict)
			| LRStateLabel s::ta => f ta n n (SE.bind1 dict (s,n))
			| (LREnEx (en,ex)::ta) => f ta (n + g(en,ex)) max dict
			| atom::ta => f ta (n+1) max dict
	    val (max,dict) = f arr 0 0 SE.empty
	in 
		if mintysize max <= tgt_size andalso 
		   (confset orelse offset >= markno) then Direct dict
		(* horrible hack: offset >= markno detects *)
		(* that we are on hand. Vulnerable!        *)
		else let val comp = fn (s1,s2) => Int.compare(#2 s1,#2 s2)
			 val lst  = Listsort.sort comp (SE.toList dict)
			 val (ids,addrs) = ListPair.unzip lst
			 val numbers = List.tabulate 
					(length ids, fn i => i + offset)
			 val dict1 = SE.bindZip SE.empty (ids,numbers)
			 val prefix = List.tabulate (offset, fn _ => 0)
		in Indirect (dict1,prefix @ addrs,max) end
	end

(* measure action and guard fields - used in transitions. NOTE: should be in
   synchro with pack_acgs function of back end *)

fun measure_acgd (sizes :lrsizes) (ac :int, gd :int) :int =
	if gd = always_ref then
		if ac = null_actref then 0
		else measure1 (#actionref sizes) (#andstref sizes)
	else measure2 (#actionref sizes, #guardref sizes) (#andstref sizes)

(* measure size of bulk of transitions *)

fun measure_trans (sizes :lrsizes) (arr :lratom list) :int =
	let val trcell  = #andstref sizes
	    val orstref = measure1 (#orstref sizes) trcell
	    val pcnc    = if (trcell >= 2) then 1 else 2
	    fun f (LRMrk          _,e) = e + 1
	      | f (LRAndStateRef  _,e) = e + 1
	      | f (LROrStateRef   _,e) = e + orstref
	      | f (LRAcGd   (ac,gd),e) = e + measure_acgd sizes (ac,gd) 
	      | f (LRCondCounters _,e) = e + pcnc 
	      | f (LRStateLabel   _,e) = e + 0 
	      | f (LREventLabel   _,e) = e + 0 
	      | f _  = impossible "#20020916a: unexpected LR opcode in trans"
	in foldl f 0 arr end

(* resolve or labels, assign addresses and compile to int lists *)

fun schedule_system (confset :bool) (sys :lrsystem) (sizes :lrsizes) :scheduling =
	let val hor_dict  = build_dict confset (#hor sys)  2      sizes (#andstref sizes)
	    val hand_dict = build_dict confset (#hand sys) markno sizes (#orstref  sizes)
	in { hor_dict  = hor_dict,
	     hand_dict = hand_dict,
	     eenv      = schedule_events (#evtypes sys),
	     tranidx   = build_tranidx (measure_trans sizes) 
				(#tenv sys) (#evtypes sys) }
	end
		
end
