(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenCstubs.sml,v
   Comment: Generate stub (faked) functions for actions (in C file)
	    and their prototypes (in header file).
   Release: scope-0-16_final 

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   CodGenCstubs.sml,v 1.11 2003/07/08 15:45:01 wasowski Exp
*)

structure CodGenCstubs :CODGEN = struct

open Absyn CodGenC 

structure S  = Symbol
structure SE = SymbolEnv
  
infix 0 << <<< <<> <- (* some fun with C++ style IO *)
infix 3 --;

(* parameter types of code generator *)

type options = bool * string option * bool (* debug, def file name, vscompatibl. *)
type aor = unit
type atr = unit

type model  = (aor,atr) Absyn.model
type system = (aor,atr) Absyn.system


val warn_generated = "/* This is a generated file! */\n/* Don't edit it manually.   */\n"


(* these are full ANSI C prototypes *)

fun mk_params [] :string = "void"
  | mk_params ps = 
	let val ts  = map tyname ps
	    val ns  = List.tabulate(length ts, fn x => " par" ^ Int.toString x)
	    val lst = ListPair.map op^ (ts,ns)
	in String.concat (join_commas lst []) end

fun dump_stubsproto (otypes :(symbol * Absyn.otype) list) :codgen = 
	let fun f ((sym,(ret,ps,_)),e) =
		e -- HDRPS ("extern " ^ tyname ret) (S.name sym) (mk_params ps)
			-- sg ";\n"
	    val e = INCLUDESTD "VSTypes.h" -- eol -- eol   
	in foldl f e otypes end

(* dump stubs for external C function. This are K&R 
   old-style functions, but proper prototypes are taken from
   header file.*)
 
fun dump_fnstubs (otypes : (symbol * Absyn.otype) list) :codgen =
	let fun f ((sym,(VoidT,ps,_)),e) =
		e -- FUNC (HDRPS (tyname VoidT) (S.name sym) (mk_params ps)) (sg "return;") 
	      | f ((sym,(ret,ps,_)),e) =
		e -- FUNC (HDRPS (tyname ret) (S.name sym) (mk_params ps)) (sg "return 1;")
	in foldl f eoc otypes end




fun dump_debugstubs (otypes : (symbol * Absyn.otype) list) :codgen =
	let fun f ((sym,(VoidT,ps,_)),e) =
		e -- FUNC (HDRPS (tyname VoidT) (S.name sym) (mk_params ps))
			(cg ["printf (\"", S.name sym, "\\n\");\n\treturn;"])
	      | f ((sym,(ret,ps,_)),e) = 
		e -- FUNC (HDRPS (tyname ret) (S.name sym) (mk_params ps))
			(cg ["printf (\"", S.name sym,"\\n\");\n\treturn 1;"])
	    val e = INCLUDESTD "stdio.h" -- eol -- eol
	in foldl f e otypes end





fun compile (debug,ofile,vscompat) ifile (model :model) :unit = 
	let val otypes = List.concat (map (SE.toList o #otypes) model)
	    val ofile = Option.getOpt (ofile,S.name(#name (hd model))^"_extern")
	    val incld = if vscompat then INCLUDESTD "SEMTypes.h" -- eol -- eol
			  else INCLUDESTD "VSTypes.h" -- eol -- eol
	    val cfile = TextIO.openOut (ofile ^ ".c")
	    val hfile = TextIO.openOut (ofile ^ ".h")
	in 
		cfile << warn_generated 
			<- incld
			<- INCLUDE (ofile ^ ".h") -- eol -- eol;

		if debug then cfile <- dump_debugstubs otypes
		else cfile  <- dump_fnstubs otypes;

		hfile << warn_generated 
			<- incld 
			<- dump_stubsproto otypes;

		TextIO.closeOut cfile;
		TextIO.closeOut hfile
	end

end

