(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Exp.sml,v
   Comment: Expression representation 
   Release: scope-0-16_final 

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Representation of expression seems to be the same throughout the project
   (the same syntax is used on input as on output, especially when targetting
   C). Theis has been factored out from Absyn to allow beter division between
   modeuls based on abstract syntax and modules based on lower level
   representations. Expressions will have common shape in all stateges though.

   Exp.sml,v 1.7 2003/03/17 15:17:57 wasowski Exp
*)

structure Exp = struct

  (* some interface *)

  open Prim Types

  type symbol = Symbol.symbol
  
  (* variables *)
  
  type var = symbol

  (* outputs aka function names *)

  type output  = symbol
  
  (* constants *)

  datatype const = CstI of int | CstR of real | CstC of char	

  (* accesses *)

  datatype access = AccessVar of var		(* x       *)
		  | AccessArr of var * exp	(* x[expr] *)

  (* expressions *)		     

  and exp = Cst    of const
	  | Access of access
	  | Prim1  of unop * exp
	  | Prim2  of binop * exp * exp
	  | Call   of oexp	    (* name, [param] *)
	  | After  of exp * exp

  (* output expression *)

  withtype oexp = output * exp list	    (* name, [param] *)	
  
  (* bexp and exp are only introduced to follow the paper more closely *)

  type bexp = exp
  type aexp = exp
  
  (* assignment *)	       
  
  type assgn = access * aexp

  datatype command = Expr  of exp 
                   | Assgn of assgn

  (* definition of purity - note that this is conservative *)
  
  fun pure_exp (Cst    c              ) = true
    | pure_exp (Prim1 (_,e)           ) = pure_exp e
    | pure_exp (Prim2 (_,e1,e2)       ) = pure_exp e1 andalso pure_exp e2
    | pure_exp (Call  _               ) = false
    | pure_exp (Access(AccessVar _)   ) = true
    | pure_exp (Access(AccessArr(_,e))) = pure_exp e
    | pure_exp (After (e1,e2)         ) = pure_exp e1 andalso pure_exp e2

  (* this function builds a sequence of expressions using After *)

  fun exp_seq []      = raise Fail "#200303170" (* should not be called *)
    | exp_seq (e::es) = foldl (fn (e1,e2)=>After(e2,e1)) e es
  
  (* constant -> value *)

  fun const (CstI i) = ValI i
    | const (CstC c) = ValI (ord c)
    | const (CstR r) = ValR r
  
  (* value -> constant *)

  fun value (ValI i) = CstI i
    | value (ValR r) = CstR r
    | value _        = raise Fail "#200302172" (* should not be called *)

end
