(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/Optimize.sml,v
   Comment: Model minimizer (or rather optimizer) for Absyn models
	    NOT FINISHED, UNSTABLE
   Release: scope-0-16_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Optimize.sml,v 1.2 2003/07/08 15:45:01 wasowski Exp
*)

structure Optimize : TRANS = struct

open Absyn

structure SE = SymbolEnv
structure S = Symbol

(* simple implementation of global DEF and USE counters *)

type symcount = int SE.env

fun addcount (n :int)  (s :S.symbol, sc :symcount) =
	let val c = SE.lookup sc s handle SE.NotFound => 0
	in SE.bind1 sc (s,c+n) end

val incr = addcount 1
val decr = addcount ~1

(* implementation of optimizer *)

type options = unit
type aor     = unit
type atr     = unit

(* optimize some basic staff on expressions *)
(* at the moment only unfold consts *)

fun opt_exp lenv cenv e :exp =
	case e of
		  Cst c => Cst c	
		| Access (AccessVar v) => 
			if List.exists (fn x=>x=v) lenv then e
			else ((Cst (SE.lookup cenv v) (*before print "#"*))
				handle SE.NotFound => e)
		| Access (AccessArr(v,e)) =>
			Access(AccessArr(v, opt_exp lenv cenv e))
		| Prim1(ope,e) => 
			Prim1(ope, opt_exp lenv cenv e)
		| Prim2(ope,e1,e2) => 
			Prim2(ope, opt_exp lenv cenv e1, 
				   opt_exp lenv cenv e2)
		| Call(ou,ps) => 
			Call(ou, map (opt_exp lenv cenv) ps)
		| After(e1,e2) => 
			After (opt_exp lenv cenv e1, 
			       opt_exp lenv cenv e2)

fun opt_command lenv cenv cmd :command option = 
	case cmd of
		  Expr e => SOME (Expr (opt_exp lenv cenv e))
		| Assgn (a,e) => SOME (Assgn (a,opt_exp lenv cenv e))

fun opt_action (lenv :var list) (cenv :const env) (cmd,sexp) :action = 
	let val cmd1 =  List.mapPartial (opt_command lenv cenv) cmd
	    val _ = assert (fn()=> List.all (fn (_,ps)=> null ps) sexp)
			"Signal parameters are not implemented!"
	in (cmd1,sexp) end

fun opt_or (cenv :const env) (ORState s) = ORState s
fun opt_and (cenv :const env) (ANDState s ) =	
	ANDState {	
		name= #name s,
		parent= #parent s,
		children= #children s,
		entry= opt_action [] cenv (#entry s),
		exit= opt_action [] cenv (#exit s)
	}


fun opt_hierarchy (cenv :const env) h :'aor hierarchy = 
	idx_hierarchy (hmapand (opt_or cenv) (opt_and cenv) (gettree h))

fun opt_tran (cenv :const env) tran :'atr tran option = 
	let val lenv = bound_names (#trigger tran)
	in SOME {
		trigger = #trigger tran,
		scope   = #scope tran,
		pos     = #pos tran,
		neg     = #neg tran,
		guard   = map (opt_exp lenv cenv) (#guard tran),
		action  = opt_action lenv cenv (#action tran),
		targets = #targets tran,
		annotation = #annotation tran }
	end



(* build initial dfa structures *)

fun init_dfa (sys as {hierarchy=h,vartypes,trans,...} :('aor,'atr) system) 
							:int env * int env =
	let val u :symcount = SE.map (fn (inter,_)=>
				if inter then 0 else 1) vartypes
	    fun add n lenv (s,sc) = if List.exists (fn x=>x=s) lenv then sc 
				    else addcount n (s,sc)
	    val inc = add 1 and dec = add ~1
	    fun fexp lenv (e,(u,d)) = 
			case e of Cst c => (u,d)
				| Access (AccessVar v) => 
					(inc lenv (v,u),d)
				| Access (AccessArr(v,e)) =>
					fexp lenv (e,(inc lenv (v,u),d))
				| Prim1 (_,e) => 
					fexp lenv (e,(u,d))
				| Prim2 (_,e1,e2) => 
					fexp lenv (e2, fexp lenv (e1,(u,d)))
				| Call (_,es) => 
					foldl (fexp lenv) (u,d) es
				| After (e1,e2) => 
					fexp lenv (e2,fexp lenv (e1,(u,d)))
	    fun fasgn lenv ((acc,e),(u,d)) = 
			let val (u1,d1) = fexp lenv (e,(u,d))
			in case acc of
				  AccessVar v => (u1,inc lenv (v,d1))
				| AccessArr(v,e) => 
					fexp lenv (e,(u1,inc lenv (v,d1)))
			end
	    fun fcmd lenv (Expr e,(u,d))  = fexp lenv (e,(u,d))
	      | fcmd lenv (Assgn a,(u,d)) = fasgn lenv (a,(u,d))
	    fun faction lenv ((cmds,_) :action, (u,d)) = 
			foldl (fcmd lenv) (u,d) cmds
	    fun fand (ANDState s,(u,d)) :symcount * symcount = 
			faction [] (#exit s,faction [] (#entry s,(u,d)))
	    fun ftran (t,(u,d)) :symcount * symcount = 
		let val localn  = bound_names (#trigger t)
		    val (u1,d1) = foldl (fexp localn) (u,d) (#guard t)
		    val (u2,d2) = faction localn (#action t,(u1,d1))
		in (u2,d2) end 
	    val (u1,d1) = hfoldand (fn (_,d)=>d) fand (u,u) (gettree h)
	    val (u2,d2) = foldl ftran (u1,d1) trans
	in (u2,d2) end

fun opt_system (sys :('aor,'atr) system) :('aor,'atr) system option =
	let val (usee,defe) = init_dfa sys
	in SOME { 
		name     = #name sys,
		fname    = #fname sys,
		vartypes = #vartypes sys,
		hierarchy= opt_hierarchy (#cenv sys) (#hierarchy sys),
		evtypes  = #evtypes sys,
		evgroups = #evgroups sys,
		otypes   = #otypes sys,
		denv     = #denv sys,
		venv     = #venv sys,
		cenv     = #cenv sys,
		trans    = List.mapPartial (opt_tran (#cenv sys)) (#trans sys)
	} end



fun apply () (mname :string) (m :('aor,'atr) model) = 
	List.mapPartial opt_system m

(* Constant propagation TBD *)

	(* This function strips outer parts of expression which
	   do not have side effects (as deeply as possible) *)

(*	fun strip_oexp (Call   c         ) = SOME (Call c)
	  | strip_oexp (Cst    c         ) = NONE
	  | strip_oexp (Prim1 (ope,e)    ) = strip_oexp e
	  | strip_oexp (Prim2 (ope,e1,e2)) = 
		if ope = OpAnd orelse ope = OpAnd then SOME (Prim2(ope,e1,e2))
		else case (strip_oexp e1, strip_oexp e2) of
			  (NONE,NONE)       => NONE
			| (SOME s1,NONE)    => SOME s1
			| (NONE,SOME s2)    => SOME s2
			| (SOME s1,SOME s2) => SOME(Prim2(OpEq,s1,s2)) *)
end
