(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/acfg/acfg-builder-fn.sml,v
   Comment: ACFG Builder (generic implementation)
   Release: scope-0-16_final

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   acfg-builder-fn.sml,v 1.7 2003/01/27 12:03:58 wasowski Exp 
*)

functor ACFGBuilderFn (G :ACFG) :> ACFG_BUILDER 
		where type (''a,''b) acfg = (''a,''b) G.acfg = 
struct

	open G
	structure B = bdd
	structure BU = BDDUtil

	val filter = List.filter
	val exists = List.exists
	val all    = List.all
	
	type ''b rule           = bdd.bdd * ''b
	type (''a,'c) splitdata = bdd.bdd * ''a * 'c * 'c
	type (''b,'c) partition = bdd.bdd * ''b rule list * 'c 
					-> (''b rule list * 'c) list
	type (''a,''b,'c) pull_up = bdd.bdd * ''b rule list * 'c 
					-> (''a,'c) splitdata
  
	(* Separates satisfied rules from pending. *)
	(* Discards rules having FALSE condition   *)
	(* (i.e. eliminated rules                  *)
	   
	fun split_rules (Fi :B.bdd) (rs :''b rule list) : ''b rule list * ''b rule list = 
		let fun f []      (s,p) = (s,p)
		      | f (r::ta) (s,p) =
				let val bdd = B.AND(Fi,#1 r)
				in if B.equal bdd B.FALSE then f ta (s,p)
				   else if B.equal bdd Fi then f ta (r::s,p) 
				   else f ta (s,r::p)
				end
		in f rs ([],[]) end



	fun build_acfg (partition :(''b,'c) partition) 
		       (pull_up :(''a,''b,'c) pull_up) Fi Lambda (e :'c) 
							:(''a,''b) acfg =

		let fun build_seq  (Fi :B.bdd, Lambda : ''b rule list, e :'c) 
							:(''a,''b) acfg = 
			let val _ = not (B.equal Fi B.FALSE) orelse raise Fail "Trying to build a sequence in \ \inaccessible branch"
			    val (sat,pending) = split_rules Fi Lambda
(*    val _ = print ( "all=" ^ Int.toString (length Lambda) ^ " matched=" ^ Int.toString (length sat) ^ " pending=" ^ Int.toString (length pending) ^ "\n" (* FIXME *)) *)
			    val trees   = if null pending then []
					  else map (build_tree Fi) 
						(partition (Fi,pending,e))
			in ACFG (map #2 sat,trees) end

		    and build_tree (Fi :B.bdd) (Lambda : ''b rule list, e :'c) 
				: (''a,''b) test = 
			let 
			    val _ = not (null Lambda) orelse 
				    raise Fail "Partition returned an empty block!"

			    val (cnd,cnd_syn,elo,ehi) = pull_up (Fi,Lambda,e)

			    val Filo = B.AND(Fi,B.NOT cnd)
			    val Fihi = B.AND(Fi,cnd)
			    
			    fun sim c (g,a) = (B.simplify g c,a)
			    val Lambdalo = map (sim Filo) Lambda
			    val Lambdahi = map (sim Fihi) Lambda

			    (* powyzsze dalo 10% speed lepiej *)
			    (* ale nie wiadomo jak po dalszych *)
			
(*			    val _ = print "FALSE " (* FIXME *)*)
			    val seqlo = build_seq (Filo, Lambdalo,elo)
(*			    val _ = print "TRUE " (* FIXME *)*)
			    val seqhi = build_seq (Fihi, Lambdahi,ehi)
(*			    val _ = print "END\n" (* FIXME *)*)
			in Test (cnd_syn, seqlo, seqhi) end
		
		in build_seq (Fi,Lambda,e) end
end
(*
	The convention for type variables:
	 ''a denotes type of conditions,
	 ''b denotes type of actions,
	 ''c denotes type of state of computation of heuristics 
*)

