(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/envspec/envtool.sml,v
   Comment: Interactive environment for environments
   Release: scope-0-16_final

     Copyright (C) 2003 Andrzej Wasowski

     portions from Moscow ML code:

     Copyright (C) Peter Sestoft
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   envtool.sml,v 1.2 2003/12/01 15:48:47 wasowski Exp
*)

app load ["Location", "Nonstdio", "envsetlex"];

open envsetabs;

(* Fancy parsing from a file; show the offending program piece on error *)

fun parseExprReport file stream lexbuf =
    let val expr = 
            envsetpar.Spec envsetlex.Token lexbuf
            handle
               Parsing.ParseError f =>
                   let val pos1 = Lexing.getLexemeStart lexbuf
                       val pos2 = Lexing.getLexemeEnd lexbuf
                   in
                       Location.errMsg (file, stream, lexbuf) 
                                       (Location.Loc(pos1, pos2))
                                       "Syntax error."
                   end
             | envsetlex.LexicalError(msg, pos1, pos2) =>
                   if pos1 >= 0 andalso pos2 >= 0 then
                       Location.errMsg (file, stream, lexbuf)
                                       (Location.Loc(pos1, pos2))
                                       ("Lexical error: " ^ msg)
                   else 
                       (Location.errPrompt ("Lexical error: " ^ msg ^ "\n\n");
                        raise Fail "Lexical error");
    in
        Parsing.clearParser();
        expr
    end
    handle exn => (Parsing.clearParser(); raise exn);

(* Parse a program from a string, with error reporting *)

fun parses str =
    parseExprReport "" (BasicIO.std_in) (Lexing.createLexerString str);

(* Create lexer from instream *)

fun createLexerStream (is : BasicIO.instream) =
    Lexing.createLexer (fn buff => fn n => Nonstdio.buff_input is buff 0 n)

(* Parse a program from a file, with error reporting *)

fun parsef file =
    let val is     = Nonstdio.open_in_bin file
        val expr   = parseExprReport file is (createLexerStream is)
                     handle exn => (BasicIO.close_in is; raise exn)
    in 
        BasicIO.close_in is;
        expr
    end

(*
(* Examples in concrete syntax *)

val pex1 = 
    parses "let f1 x = x + 1 in f1 12 end";


(* Example: factorial *)

val pex2 = 
    parses "let fac x = if x=0 then 1 else x * fac (x-1) in fac n end";

(* Example: deep recursion to check for constant-space tail recursion *)

val pex3 = 
    parses "let deep x = if x=0 then 1 else deep (x-1) in deep count end";

val ok = [ex1, ex2, ex2] = [pex1, pex2, pex2];
*)
