(* envtooli.sml,v 1.3 2003/12/01 22:32:40 wasowski Exp *)

structure envtooli :> ENVTOOLI = struct

  open Lexing Parsing Err Say envsetpar

  structure L = Location
  exception LexicalError = envsetlex.LexicalError
  exception Toplevel
  exception EndOfFile

  (* Create lexer from file (given by path) *)

  fun createLexerStream (is : BasicIO.instream) =
    Lexing.createLexer (fn buff => fn n => Nonstdio.buff_input is buff 0 n)

  val input_lexbuf = ref (createLexerStream BasicIO.std_in)



 fun parsePhrase parsingFun lexingFun lexbuf =
  let fun skip() =
    (case lexingFun lexbuf of
        T_eof  => ()
      | T_semi => ()
      | _ => skip())
    handle LexicalError(_,_,_) =>
      skip()
  in
    parsingFun lexingFun lexbuf
    handle
        Parsing.ParseError f =>
           let val pos1 = Lexing.getLexemeStart lexbuf
               val pos2 = Lexing.getLexemeEnd lexbuf
           in
             (*if f (Obj.repr T_eof) orelse f (Obj.repr T_semi)
             then () else skip();*)
             print "Syntax error.\n";
             raise Toplevel
           end
       | LexicalError(msg, pos1, pos2) =>
	   (print ("Lexical error: " ^ msg);
            skip();
            raise Toplevel)
       | Toplevel =>
           (skip ();
            raise Toplevel)
  end ;

  fun parsePhraseAndClear parsingFun lexingFun lexbuf =
  let val phr =
    parsePhrase parsingFun lexingFun lexbuf
    handle x => (Parsing.clearParser(); raise x)
  in
    Parsing.clearParser();
    phr
  end

  val parseToplevelPhrase =
	parsePhraseAndClear envsetpar.SpecPhrase envsetlex.Token

  fun execToplevelPhrase phrase = () (* TODO *)

  fun loadToplevelPhrase lexbuf =
	let val (phrase, isLast) = parseToplevelPhrase lexbuf 
	in execToplevelPhrase phrase; isLast end

  fun eval_loop () =
  while true do (
      print "- ";
      let val isLast = loadToplevelPhrase (!input_lexbuf)
      in if isLast then raise EndOfFile else () end 
      handle
          EndOfFile => (print "\n"; BasicIO.exit 0)
        | Toplevel => BasicIO.flush_out BasicIO.std_out
        | Interrupt => print "Interrupted.\n"
        | x => (BasicIO.flush_out BasicIO.std_out; raise x)
  );

  val _ = print ("envtool " ^ rcs.release ^ " [envset mode] " ^ rcs.date ^ "\n")
  val _ = eval_loop()

end