(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/util/env-sig.sml,v 
   Comment: Interface to environments inspired by Peter Sestoft in PFOO
            course at IT-Universty
   Release: scope-0-16_final

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   The interface is similar to one proposed by Peter. Only key is no longer a
   type variable (of equality type), but a fixed type (so it is dropped from
   many parameters). key may be any type for which toInt function may be
   given to the functor. 

   env-sig.sml,v 1.5 2003/03/14 12:36:39 wasowski Exp
*)   

signature ENV = sig

  type key
  type 'data env

  exception NotFound 
  
  val empty    : 'data env
  val isEmpty  : 'data env -> bool
  val lookup   : 'data env -> key -> 'data		(* raise NotFound *)
  val count    : 'data env -> int
  val defined  : 'data env -> key -> bool

  val bind1    : 'data env -> key * 'data -> 'data env
  val bind     : 'data env -> (key * 'data) list -> 'data env
  val bindZip  : 'data env -> key list * 'data list -> 'data env
  
  val remove1  : 'data env -> key      -> 'data env	(* raise NotFound *)
  val remove   : 'data env -> key list -> 'data env 

  val fromList : (key * 'data) list -> 'data env
  val toList   : 'data env -> (key * 'data) list  (* for test purposes *)

  val plus     : 'data env * 'data env -> 'data env
  val join     : ('data * 'data -> 'data) -> 'data env * 'data env -> 'data env
  val minus    : 'data env * key list  -> 'data env
  
  val map      : ('a -> 'b) -> 'a env -> 'b env
  val fullmap  : (key * 'a -> 'b) -> 'a env -> 'b env
  val filter   : (key -> bool) -> 'data env -> 'data env

end  

(* 
	[remove1 env k] removes entry with key k from env. Raises NotFound
	if k is not in the domain of env.

	[remove env ks] removes all entries with keys on the ks list from
	env. Raises NotFound if any of k in ks is not in the domain of env.
	
	[minus (env,ks)] Like remove but does not raise any exceptions
	(silently ignores keys non existent in env).

	[plus  (env1,env2)] Adds all entries from env2 to env1, overwriting
	new entries

	[join f (env1,env2)] This is a generalized version of plus. It adds
	to environments like plus, but instead of overwriting the overlapping
	entries in env2 it joins the entries into one using function f.

	[filter p env] restrict environment to only contain bindings
	satisfying p, where p is a boolean predicate over keys.
*)


